##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""

We create an app:

  >>> getRootFolder()["app"] = App()

We set up a test browser:

  >>> from zope.testbrowser.testing import Browser
  >>> browser = Browser()
  >>> browser.addHeader('Authorization', 'Basic mgr:mgrpw')
  >>> browser.handleErrors = False

Let's look at a department::

  >>> browser.open("http://localhost/app/departments/1")
  >>> print browser.contents
  Department: 1

We can also go to an employee::

  >>> browser.open("http://localhost/app/departments/1/employees/2")
  >>> print browser.contents
  Employee: 1 2

We cannot go to departments as there's no view for it::

  >>> browser.open("http://localhost/app/departments")
  Traceback (most recent call last):
    ...
  NotFound: Object: <megrok.traject.components.DefaultModel object at ...>,
  name: u'@@index'
    
Nor can we go to employees::

  >>> browser.open("http://localhost/app/departments/1/employees")
  Traceback (most recent call last):
    ...
  NotFound: Object: <megrok.traject.components.DefaultModel object at ...>,
  name: u'@@index'
"""

import grok

from megrok import traject

class Department(grok.Model):
    def __init__(self, department_id):
        self.department_id = department_id

class Employee(grok.Model):
    def __init__(self, department_id, employee_id):
        self.department_id = department_id
        self.employee_id = employee_id

class DepartmentIndex(grok.View):
    grok.context(Department)
    grok.name('index')
    def render(self):
        return 'Department: %s' % self.context.department_id

class EmployeeIndex(grok.View):
    grok.context(Employee)
    grok.name('index')
    def render(self):
        return 'Employee: %s %s' % (self.context.department_id,
                                    self.context.employee_id)

class App(grok.Application, grok.Model):
    pass

class DepartmentTraject(traject.Traject):
    grok.context(App)

    pattern = 'departments/:department_id'
    model = Department

    def factory(department_id):
        return Department(department_id)

    def arguments(department):
        return dict(department_id=department.department_id)

class EmployeeTraject(traject.Traject):
    grok.context(App)

    pattern = 'departments/:department_id/employees/:employee_id'
    model = Employee

    def factory(department_id, employee_id):
        return Employee(department_id, employee_id)

    def arguments(employee):
        return dict(
            department_id=employee.department_id,
            employee_id=employee.employee_id)



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""

We create an app:

  >>> getRootFolder()["app"] = App()

We set up a test browser:

  >>> from zope.testbrowser.testing import Browser
  >>> browser = Browser()
  >>> browser.addHeader('Authorization', 'Basic mgr:mgrpw')
  >>> browser.handleErrors = False

We are going to look at the request.getURL() behavior of various
views. We need to get these right otherwise things break in a weird
way.

App::

  >>> browser.open("http://localhost/app")
  >>> print browser.contents
  http://localhost/app/@@index

App explicit index::

  >>> browser.open("http://localhost/app/@@index")
  >>> print browser.contents
  http://localhost/app/@@index

App explicit index without @@::

  >>> browser.open("http://localhost/app/index")
  >>> print browser.contents
  http://localhost/app/index

Departments::

  >>> browser.open("http://localhost/app/departments")
  >>> print browser.contents
  http://localhost/app/departments/@@index

Departments explicit index::

  >>> browser.open("http://localhost/app/departments/@@index")
  >>> print browser.contents
  http://localhost/app/departments/@@index

Departments explicit index without @@::

  >>> browser.open("http://localhost/app/departments/index")
  >>> print browser.contents
  http://localhost/app/departments/index

Department::

  >>> browser.open("http://localhost/app/departments/1")
  >>> print browser.contents
  http://localhost/app/departments/1/@@index

Department explicit index::

  >>> browser.open("http://localhost/app/departments/1/@@index")
  >>> print browser.contents
  http://localhost/app/departments/1/@@index

Department explicit index without @@::

  >>> browser.open("http://localhost/app/departments/1/index")
  >>> print browser.contents
  http://localhost/app/departments/1/index

Employees::

  >>> browser.open("http://localhost/app/departments/1/employees")
  >>> print browser.contents
  http://localhost/app/departments/1/employees/@@index

Employees explicit index::

  >>> browser.open("http://localhost/app/departments/1/employees/@@index")
  >>> print browser.contents
  http://localhost/app/departments/1/employees/@@index

Employees explicit index without @@::

  >>> browser.open("http://localhost/app/departments/1/employees/index")
  >>> print browser.contents
  http://localhost/app/departments/1/employees/index
  
Employee::

  >>> browser.open("http://localhost/app/departments/1/employees/2")
  >>> print browser.contents
  http://localhost/app/departments/1/employees/2/@@index

Employee explicit index::

  >>> browser.open("http://localhost/app/departments/1/employees/2/@@index")
  >>> print browser.contents
  http://localhost/app/departments/1/employees/2/@@index

Employee explicit index without @@::

  >>> browser.open("http://localhost/app/departments/1/employees/2/index")
  >>> print browser.contents
  http://localhost/app/departments/1/employees/2/index
  
"""

import grok

from megrok import traject

class Departments(grok.Model):
    pass

class Department(grok.Model):
    def __init__(self, department_id):
        self.department_id = department_id

class Employees(grok.Model):
    def __init__(self, department_id):
        self.department_id = department_id
    
class Employee(grok.Model):
    def __init__(self, department_id, employee_id):
        self.department_id = department_id
        self.employee_id = employee_id

class DepartmentsIndex(grok.View):
    grok.context(Departments)
    grok.name('index')
    def render(self):
        return self.request.getURL()

class DepartmentIndex(grok.View):
    grok.context(Department)
    grok.name('index')
    def render(self):
        return self.request.getURL()

class EmployeesIndex(grok.View):
    grok.context(Employees)
    grok.name('index')
    def render(self):
        return self.request.getURL()

class EmployeeIndex(grok.View):
    grok.context(Employee)
    grok.name('index')
    def render(self):
        return self.request.getURL()

class App(grok.Application, grok.Model):
    pass

class AppIndex(grok.View):
    grok.context(App)
    grok.name('index')
    def render(self):
        return self.request.getURL()
    
class DepartmentsTraject(traject.Traject):
    grok.context(App)

    pattern = 'departments'
    model = Departments

    def factory():
        return Departments()

    def arguments(department):
        return {}

class DepartmentTraject(traject.Traject):
    grok.context(App)

    pattern = 'departments/:department_id'
    model = Department

    def factory(department_id):
        try:
            department_id = int(department_id)
        except ValueError:
            return None
        return Department(department_id)

    def arguments(department):
        return dict(department_id=department.department_id)

class EmployeesTraject(traject.Traject):
    grok.context(App)

    pattern = 'departments/:department_id/employees'
    model = Employees

    def factory(department_id):
        try:
            department_id = int(department_id)
        except ValueError:
            return None
        return Employees(department_id)

    def arguments(employees):
        return dict(department_id=employees.department_id)

class EmployeeTraject(traject.Traject):
    grok.context(App)

    pattern = 'departments/:department_id/employees/:employee_id'
    model = Employee

    def factory(department_id, employee_id):
        try:
            department_id = int(department_id)
            employee_id = int(employee_id)
        except ValueError:
            return None
        return Employee(department_id, employee_id)

    def arguments(employee):
        return dict(
            department_id=employee.department_id,
            employee_id=employee.employee_id)



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

"""megrok.traject components"""

import grok
import traject
from grokcore.component import Context

class Traject(object):
    grok.baseclass()

    pattern = None
    model = None
    
    def factory(**kw):
        raise NotImplementedError

    def arguments(obj):
        return NotImplementedError

class TrajectTraverser(grok.Traverser):
    grok.baseclass()
        
    def traverse(self, name):
        stack = self.request.getTraversalStack()
        stack.append(name)
        unconsumed, consumed, obj = traject.consume_stack(
            self.context, stack, DefaultModel)
        # if we haven't gone anywhere, we can't traverse; fall back
        if obj is self.context:
            return None
        self.request.setTraversalStack(unconsumed)
        # remove the first consumed element, as it will actually be placed
        # in _traversed_names already
        consumed = consumed[1:]
        # we need to mess around with _traversed_names otherwise
        # the request.getURL and base URL generation breaks
        self.request._traversed_names.extend(consumed)

        return obj

class DefaultModel(grok.Model):
    def __init__(self, **kw):
        self.kw = kw

class Model(Context):
    pass



import martian

class pattern(martian.Directive):
    scope = martian.CLASS
    store = martian.ONCE
    validate = martian.validateText

# these are dummy decorators for now
def factory(object):
    return object

def arguments(object):
    return object




##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""

We create an app:

  >>> getRootFolder()["app"] = app = App()

We set up a test browser:

  >>> from zope.testbrowser.testing import Browser
  >>> browser = Browser()
  >>> browser.addHeader('Authorization', 'Basic mgr:mgrpw')
  >>> browser.handleErrors = False

We we traverse from the app URL, we expect things to happen::

  >>> browser.open("http://localhost/app/mammoths/Knuth")
  >>> print browser.contents
  The name of this mammoth is Knuth.

We can also go to another view::

  >>> browser.open("http://localhost/app/mammoths/Knuth/other")
  >>> print browser.contents
  This is indeed Knuth

We can also locate a mammoth::

  >>> mammoth = Mammoth('Dijkstra')
  >>> class Default(object):
  ...    def __init__(self, *args, **kw):
  ...        pass
  >>> traject.locate(app, mammoth, Default)
  >>> mammoth.__name__
  u'Dijkstra'
  >>> mammoth.__parent__
  <megrok.traject.ftests.traject.model.Default object at ...>

"""

import grok

from megrok import traject

class App(grok.Application, grok.Model):
    pass

class Mammoth(traject.Model):
    traject.context(App)
    traject.pattern('mammoths/:name')
    
    def __init__(self, name):
        self.name = name

    def factory(name):
        return Mammoth(name)

    def arguments(self):
        return dict(name=self.name)

class MammothIndex(grok.View):
    grok.context(Mammoth)
    grok.name('index')
    def render(self):
        return 'The name of this mammoth is %s.' % self.context.name

class MammothOther(grok.View):
    grok.context(Mammoth)
    grok.name('other')
    def render(self):
        return "This is indeed %s" % self.context.name



import re
import unittest
import megrok.traject
import os

from pkg_resources import resource_listdir
from zope.testing import doctest, renormalizing
from zope.app.testing.functional import (HTTPCaller, getRootFolder,
                                         FunctionalTestSetup, sync, ZCMLLayer)

ftesting_zcml = os.path.join(os.path.dirname(megrok.traject.__file__),
                             'ftesting.zcml')
FunctionalLayer = ZCMLLayer(ftesting_zcml, __name__, 'FunctionalLayer')

def setUp(test):
    FunctionalTestSetup().setUp()

def tearDown(test):
    FunctionalTestSetup().tearDown()

checker = renormalizing.RENormalizing([
    # Accommodate to exception wrapping in newer versions of mechanize
    (re.compile(r'httperror_seek_wrapper:', re.M), 'HTTPError:'),
    ])

def suiteFromPackage(name):
    files = resource_listdir(__name__, name)
    suite = unittest.TestSuite()
    for filename in files:
        if not filename.endswith('.py'):
            continue
        if filename == '__init__.py':
            continue

        dottedname = 'megrok.traject.ftests.%s.%s' % (name, filename[:-3])
        test = doctest.DocTestSuite(
            dottedname, setUp=setUp, tearDown=tearDown,
            checker=checker,
            extraglobs=dict(http=HTTPCaller(),
                            getRootFolder=getRootFolder,
                            sync=sync),
            optionflags=(doctest.ELLIPSIS+
                         doctest.NORMALIZE_WHITESPACE+
                         doctest.REPORT_NDIFF)
            )
        test.layer = FunctionalLayer

        suite.addTest(test)
    return suite

def test_suite():
    suite = unittest.TestSuite()
    for name in ['traject']:
        suite.addTest(suiteFromPackage(name))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""

We create an app:

  >>> getRootFolder()["app"] = App()

We set up a test browser:

  >>> from zope.testbrowser.testing import Browser
  >>> browser = Browser()
  >>> browser.addHeader('Authorization', 'Basic mgr:mgrpw')

When we traverse to an object that doesn't exist, we expect an error::

  >>> browser.open("http://localhost/app/doesntexist")
  Traceback (most recent call last):
     ...
  httperror_seek_wrapper: HTTP Error 404: Not Found

"""

import grok

from megrok import traject

class Mammoth(grok.Model):
    def __init__(self, name):
        self.name = name

class MammothIndex(grok.View):
    grok.context(Mammoth)
    grok.name('index')
    def render(self):
        return 'The name of this mammoth is %s.' % self.context.name

class App(grok.Application, grok.Model):
    pass

class MammothTraject(traject.Traject):
    grok.context(App)

    pattern = 'mammoths/:name'
    model = Mammoth

    def factory(name):
        return Mammoth(name)

    def arguments(mammoth):
        return dict(name=mammoth.name)



import os
from setuptools import setup, find_packages

def read(*filenames):
    return open(os.path.join(os.path.dirname(__file__), *filenames)).read()

long_description = (
    read('README.txt')
    + '\n' +
    read('CHANGES.txt')
    + '\n' +
    'Download\n'
    '********\n'
    )

setup(name='megrok.traject',
      version='1.0',
      description="Traject integration for Grok applications",
      long_description=long_description,
      # Use classifiers that are already listed at:
      # http://pypi.python.org/pypi?%3Aaction=list_classifiers
      classifiers=['Development Status :: 4 - Beta',
                   'Environment :: Web Environment',
                   'Framework :: Zope3',
                   'Intended Audience :: Developers',
                   'License :: OSI Approved :: Zope Public License',
                   'Operating System :: OS Independent',
                   'Programming Language :: Python',
                   'Topic :: Internet :: WWW/HTTP',
                   'Topic :: Software Development :: Libraries',
                   ],
      keywords="grok megrok url traject traversal routing route",
      author="Martijn Faassen",
      author_email="faassen@startifact.com",
      license="ZPL",
      package_dir={'': 'src'},
      namespace_packages=['megrok'],
      packages=find_packages('src'),
      include_package_data=True,
      zip_safe=False,
      install_requires=['setuptools',
                        'grok', # for grok.Traverser, ideally get rid of that
                        'traject',
                        'martian',
                        'grokcore.component',
                        'zope.publisher',
                        'zope.component',
                        ],
      entry_points="""
      # Add entry points here
      """,
      )



import traject
import martian
from megrok.traject import components, directive
import grokcore.component
from zope import component
from zope.publisher.interfaces.browser import (IBrowserPublisher)
from zope.publisher.interfaces.http import IHTTPRequest

class TrajectGrokker(martian.ClassGrokker):
    martian.component(components.Traject)

    martian.directive(grokcore.component.context)

    def execute(self, factory, config, context, **kw):
        pattern_str = factory.pattern
        model = factory.model

        factory_func = _get_func(factory, 'factory')
        arguments_func = _get_func(factory, 'arguments')

        _register_traject(config, context, pattern_str, model,
                          factory_func, arguments_func)
        return True

class ModelGrokker(martian.ClassGrokker):
    martian.component(components.Model)

    martian.directive(grokcore.component.context)
    martian.directive(directive.pattern)
    
    def execute(self, factory, config, context, pattern, **kw):
        pattern_str = pattern
        model = factory

        factory_func = _get_func(factory, 'factory')
        arguments_func = _get_func(factory, 'arguments')

        _register_traject(config, context, pattern_str, model,
                          factory_func, arguments_func)
        return True

def _get_func(cls, name):
    f = getattr(cls, name)
    # detect a @classmethod (XXX is this correct?)
    if f.im_self is not None:
        return f
    # return function underlying method implementation
    return f.im_func

def _register_traject(config, context,
                      pattern_str, model, factory_func, arguments_func):
    # register
    config.action(
        discriminator=('traject.register', context, pattern_str),
        callable=traject.register,
        args=(context, pattern_str, factory_func),
        )

    # register_inverse
    config.action(
        discriminator=('traject.register_inverse', context, model,
                       pattern_str),
        callable=traject.register_inverse,
        args=(context, model, pattern_str, arguments_func))
    
    # register traverser on context; overwrite previous as they're
    # all the same
    config.action(
        discriminator=object(), # we always want a different discriminator
        callable=component.provideAdapter,
        args=(components.TrajectTraverser, (context, IHTTPRequest),
              IBrowserPublisher),
        )
